#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2019 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the 'License'); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import hashlib
import hmac
import logging
import json

from urllib.parse import unquote
from urllib.parse import quote

from dciauth.v2.time import get_now

logger = logging.getLogger(__name__)

TIMESTAMP_FORMAT = "%Y%m%dT%H%M%SZ"
DATESTAMP_FORMAT = "%Y%m%d"


def encode_data(data):
    try:
        return (data or "").encode("utf-8")
    except (UnicodeDecodeError, AttributeError):
        return data


def generate_headers(request, credentials):
    access_key = credentials.get("access_key")
    secret_key = credentials.get("secret_key")
    logger.debug("Generate HMAC v2 headers for %s" % access_key)
    if not access_key or not secret_key:
        return {}
    if "payload" in request:
        payload = request.pop("payload")
        request["data"] = json.dumps(payload)
    request["data"] = encode_data(request.get("data"))
    now = get_now()
    request["timestamp"] = request.get("timestamp", now.strftime(TIMESTAMP_FORMAT))
    request["datestamp"] = request.get("datestamp", now.strftime(DATESTAMP_FORMAT))
    date_header = request.get("date_header", "X-DCI-Date")
    headers = {
        date_header: request["timestamp"],
        "Authorization": _build_authorization_header(request, access_key, secret_key),
    }
    logger.debug("Generated headers %s" % json.dumps(headers, indent=2, sort_keys=True))
    return headers


def _build_authorization_header(request, access_key, secret_key):
    string_to_sign = _get_string_to_sign(request)
    signing_key = _get_signing_key(request, secret_key)
    signature = hmac.new(signing_key, string_to_sign, hashlib.sha256).hexdigest()
    # pylint: disable=line-too-long
    return """{algorithm} Credential={access_key}/{credential_scope}, SignedHeaders={signed_headers}, Signature={signature}""".format(
        algorithm=_get_algorithm(request),
        access_key=access_key,
        credential_scope=_get_credential_scope(request),
        signed_headers=_get_signed_headers(request),
        signature=signature,
    )
    # pylint: enable=line-too-long


def _get_string_to_sign(request):
    string_to_sign = """{algorithm}
{timestamp}
{credential_scope}
{canonical_request}""".format(
        algorithm=_get_algorithm(request),
        timestamp=request["timestamp"],
        credential_scope=_get_credential_scope(request),
        canonical_request=_get_canonical_request(request),
    )
    logger.debug("String to sign %s" % string_to_sign)
    return string_to_sign.encode("utf-8")


def _get_canonical_request(request):
    canonical_request = """{method}
{endpoint}
{canonical_querystring}
{canonical_headers}
{signed_headers}
{payload_hash}""".format(
        method=request.get("method", "GET"),
        endpoint=_get_endpoint(request),
        canonical_querystring=_get_canonical_querystring(request),
        canonical_headers=_get_canonical_headers(request),
        signed_headers=_get_signed_headers(request),
        payload_hash=_get_payload_hash(request),
    )
    logger.debug("Canonical request %s" % canonical_request)
    return hashlib.sha256(canonical_request.encode("utf-8")).hexdigest()


def _get_endpoint(request):
    return unquote(request.get("endpoint", "/"))


def _get_canonical_querystring(request):
    params = request.get("params")
    if not params:
        return ""

    key_val_pairs = []
    for key, value in params.items():
        key_val_pairs.append((quote(key, safe="_.-~"), quote(str(value), safe="_.-~")))

    sorted_key_vals_pairs = []
    for key, value in sorted(key_val_pairs):
        sorted_key_vals_pairs.append(f"{key}={value}")
    return "&".join(sorted_key_vals_pairs)


def _get_payload_hash(request):
    data = request.get("data")
    return hashlib.sha256(data).hexdigest()


def _get_canonical_headers(request):
    canonical_headers = request.get(
        "canonical_headers",
        {
            "host": request.get("host", "api.distributed-ci.io"),
            "x-dci-date": request["timestamp"],
        },
    )
    signed_headers = _get_signed_headers(request)
    return (
        "\n".join(
            ["%s:%s" % (h, canonical_headers[h]) for h in signed_headers.split(";")]
        )
        + "\n"
    )


def _get_credential_scope(request):
    return """{datestamp}/{region}/{service}/{request_type}""".format(
        datestamp=request["datestamp"],
        region=_get_region(request),
        service=_get_service(request),
        request_type=_get_request_type(request),
    )


def _get_signing_key(request, key):
    algorithm = _get_algorithm(request)
    algo_version = algorithm.replace("-HMAC-SHA256", "")
    datestamp = request["datestamp"]
    key_date = _sign((algo_version + key).encode("utf-8"), datestamp)
    region = _get_region(request)
    key_region = _sign(key_date, region)
    service = _get_service(request)
    key_service = _sign(key_region, service)
    request_type = _get_request_type(request)
    return _sign(key_service, request_type)


def _sign(key, msg):
    return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()


def _get_signed_headers(request):
    return request.get("signed_headers", "host;x-dci-date")


def _get_algorithm(request):
    return request.get("algorithm", "DCI2-HMAC-SHA256")


def _get_region(request):
    return request.get("region", "BHS3")


def _get_service(request):
    return request.get("service", "api")


def _get_request_type(request):
    return request.get("request_type", "dci2_request")


def _lower(headers):
    return {key.lower(): value for key, value in headers.items()}


def parse_headers(headers):
    headers = _lower(headers)
    timestamp = _parse_timestamp(headers)
    authorization = headers.get("authorization")
    if not timestamp or not authorization:
        return None
    algorithm, credential, signed_headers, signature = authorization.split(" ")
    signature = signature.replace("Signature=", "")
    credential = _find_in_str_between(credential, "Credential=", ",").split("/")
    if len(credential) != 6:
        return None
    signed_headers = _find_in_str_between(signed_headers, "SignedHeaders=", ",")
    request_type = credential[5]
    date_headers = {
        "aws4_request": "X-Amz-Date",
        "dci2_request": "X-DCI-Date",
    }
    if request_type not in date_headers.keys():
        return None
    parsed_headers = {
        "host": headers.get("host"),
        "algorithm": algorithm,
        "client_type": credential[0],
        "client_id": credential[1],
        "datestamp": credential[2],
        "date_header": date_headers[request_type],
        "region": credential[3],
        "service": credential[4],
        "request_type": request_type,
        "signed_headers": signed_headers,
        "canonical_headers": {h: headers[h] for h in signed_headers.split(";")},
        "timestamp": timestamp,
        "signature": signature,
    }
    logger.debug(
        "Parsed headers %s" % json.dumps(parsed_headers, indent=2, sort_keys=True)
    )
    return parsed_headers


def _parse_timestamp(headers):
    aws_date_header = "x-amz-date"
    dci_date_header = "x-dci-date"
    if aws_date_header not in headers and dci_date_header not in headers:
        return None
    return (
        headers[aws_date_header]
        if aws_date_header in headers
        else headers[dci_date_header]
    )


def _find_in_str_between(string, first, last):
    try:
        start = string.index(first) + len(first)
        end = string.index(last, start)
        return string[start:end]
    except ValueError:
        return ""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2019 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the 'License'); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import hmac
from datetime import datetime, timedelta

from dciauth.v2.headers import generate_headers, TIMESTAMP_FORMAT, _lower
from dciauth.v2.time import get_now


def is_valid(request, credential, parsed_headers):
    if parsed_headers is None:
        return False, "headers are malformated"

    if is_expired(request, parsed_headers):
        return False, "signature is expired"

    claimed_request = _get_claimed_request(request, parsed_headers)
    claimed_credential = _get_claimed_credential(credential, parsed_headers)
    claimed_headers = generate_headers(claimed_request, claimed_credential)
    signature = parsed_headers.get("signature")
    claimed_signature = _get_signature(_lower(claimed_headers))
    if _signature_equals(signature, claimed_signature):
        return True, ""

    return False, "signature is invalid"


def is_expired(request, parsed_headers):
    timestamp = parsed_headers["timestamp"]
    if timestamp:
        timestamp = datetime.strptime(timestamp, TIMESTAMP_FORMAT)
        now = get_now()
        fifteen_min = timedelta(minutes=15)
        return abs(now - timestamp) > fifteen_min
    return True


def _get_claimed_request(request, headers):
    claimed_request = headers
    claimed_request.update(
        {
            "method": request.get("method", "GET"),
            "endpoint": request.get("endpoint", "/"),
            "params": request.get("params", {}),
            "data": request.get("data", ""),
        }
    )
    return claimed_request


def _get_claimed_credential(credential, headers):
    return {
        "access_key": "%s/%s" % (headers["client_type"], headers["client_id"]),
        "secret_key": credential["secret_key"],
    }


def _get_signature(headers):
    kv_signature = headers.get("authorization").split(" ")[3]
    signature = kv_signature.split("Signature=")[1]
    return signature


def _signature_equals(signature1, signature2):
    return hmac.compare_digest(signature1.encode("utf-8"), signature2.encode("utf-8"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2017-2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the 'License'); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os
import setuptools
import subprocess

# dcibuild can be loaded only when doing the sdist sub-command because
# dci-packaging is extracted at the same level. When doing the other
# sub-commands like build, we extract the version from version.py.
try:
    from dcibuild import sdist, get_version

    sdist.dci_mod = "dciauth"
except:
    sdist = None

    def get_version():
        from dciauth import version

        return version.__version__


root_dir = os.path.dirname(os.path.abspath(__file__))
readme = open(os.path.join(root_dir, "README.md")).read()


setuptools.setup(
    name="dciauth",
    version=get_version(),
    packages=["dciauth", "dciauth.v2"],
    author="Distributed CI team",
    author_email="distributed-ci@redhat.com",
    description="DCI authentication module used by dci-control-server and python-dciclient",
    long_description=readme,
    long_description_content_type="text/markdown",
    install_requires=[],
    url="https://github.com/redhat-cip/python-dciauth",
    license="Apache v2.0",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Security :: Cryptography",
    ],
    cmdclass={
        "sdist": sdist,
    },
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2017-2024 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the 'License'); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import functools
import hashlib
import hmac
import json
import logging

from datetime import datetime, timedelta, timezone

from requests.auth import AuthBase
from urllib.parse import unquote
from urllib.parse import quote
from urllib.parse import parse_qsl
from urllib.parse import urlparse

logger = logging.getLogger(__name__)


TIMESTAMP_FORMAT = "%Y%m%dT%H%M%SZ"


class BaseHmacSignature:
    def __init__(self, options, now=None):
        self.now = datetime.now(timezone.utc) if now is None else now
        self.error_message = ""
        self.algorithm = options["algorithm"]
        self.service_name = options["service_name"]
        self.service_key = options["service_key"]
        self.region_name = options["region_name"]
        date_headers = {
            "aws4_request": "x-amz-date",
            "dci2_request": "x-dci-date",
        }
        self.date_header = date_headers[self.service_key]

    def add_request(self, request):
        logger.debug("Calculating signature using v4 auth.")
        self.parse_request(request)
        self.canonical_request = self._get_canonical_request()
        logger.debug(f"Canonical request:\n{self.canonical_request}")
        self.string_to_sign = self._get_string_to_sign()
        logger.debug(f"String to sign:\n{self.string_to_sign}")
        return self

    def parse_request(self, request):
        raise NotImplementedError

    def _get_canonical_querystring(self, args):
        key_val_pairs = []
        for key, value in args.items():
            key_val_pairs.append(
                (quote(key, safe="_.-~"), quote(str(value), safe="_.-~"))
            )

        sorted_key_vals_pairs = []
        for key, value in sorted(key_val_pairs):
            sorted_key_vals_pairs.append(f"{key}={value}")
        canonical_query_string = "&".join(sorted_key_vals_pairs)
        return canonical_query_string

    def _get_canonical_headers(self, headers_to_sign):
        headers = []
        for key in sorted(headers_to_sign):
            value = " ".join(headers_to_sign[key].split())
            headers.append(f"{key}:{value}")
        return "\n".join(headers)

    def _get_signed_headers(self, headers_to_sign):
        headers = sorted(headers_to_sign)
        return ";".join(headers)

    def _encode_data(self, data):
        try:
            return (data or "").encode("utf-8")
        except (UnicodeDecodeError, AttributeError):
            return data

    def _get_payload(self, data):
        data = self._encode_data(data)
        if data and hasattr(data, "seek"):
            position = data.tell()
            performant_payload_buffer = 1024 * 1024
            read_chunksize = functools.partial(data.read, performant_payload_buffer)
            checksum = hashlib.sha256()
            for chunk in iter(read_chunksize, b""):
                checksum.update(chunk)
            hex_checksum = checksum.hexdigest()
            data.seek(position)
            return hex_checksum
        return hashlib.sha256(data).hexdigest()

    def _get_canonical_request(self):
        canonical_request = [
            self.method,
            unquote(self.path),
            self._get_canonical_querystring(self.params),
            self._get_canonical_headers(self.headers_to_sign) + "\n",
            self._get_signed_headers(self.headers_to_sign),
            self._get_payload(self.data),
        ]
        return "\n".join(canonical_request)

    def _get_credential_scope(self):
        credential_scope = [
            self.timestamp[0:8],
            self.region_name,
            self.service_name,
            self.service_key,
        ]
        return "/".join(credential_scope)

    def _hash_sha256(self, msg):
        return hashlib.sha256(msg.encode("utf-8")).hexdigest()

    def _get_string_to_sign(self):
        string_to_sign = [
            self.algorithm,
            self.timestamp,
            self._get_credential_scope(),
            self._hash_sha256(self.canonical_request),
        ]
        return "\n".join(string_to_sign)

    def _sign_hex(self, key, msg):
        return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).hexdigest()

    def _sign(self, key, msg):
        return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()

    def _get_signature(self, secret_key):
        algo_version = self.algorithm.replace("-HMAC-SHA256", "")
        k_date = self._sign(
            (f"{algo_version}{secret_key}").encode("utf-8"), self.timestamp[0:8]
        )
        k_region = self._sign(k_date, self.region_name)
        k_service = self._sign(k_region, self.service_name)
        k_signing = self._sign(k_service, self.service_key)
        return self._sign_hex(k_signing, self.string_to_sign)

    def _signature_equals(self, signature1, signature2):
        return hmac.compare_digest(
            signature1.encode("utf-8"), signature2.encode("utf-8")
        )


class FlaskHmacSignature(BaseHmacSignature):
    def parse_request(self, request):
        self.parse_headers(request.headers)
        self.method = request.method.upper()
        self.path = request.path
        self.params = request.args
        self.data = request.data

    def parse_headers(self, headers):
        logger.debug(f"Request headers:\n{headers}")
        lower_headers = {key.lower(): value for key, value in headers.items()}
        for key, value in lower_headers.items():
            value = value.strip()
            if key == self.date_header:
                self.timestamp = value
            if key == "authorization":
                _, credential, signed_headers, signature = value.split(" ")
                credential = (
                    credential.replace("Credential=", "").replace(",", "").split("/")
                )
                if len(credential) < 5:
                    raise Exception("Credential invalid")
                self.access_key = "/".join(credential[:-4])
                signed_headers = [
                    h.lower().strip()
                    for h in signed_headers.replace("SignedHeaders=", "")
                    .replace(",", "")
                    .split(";")
                ]
                self.headers_to_sign = {
                    h: lower_headers[h] for h in set(signed_headers)
                }
                self.signature_in_header = signature.replace("Signature=", "")

    def is_expired(self):
        if self.timestamp:
            timestamp = datetime.strptime(self.timestamp, TIMESTAMP_FORMAT)
            timestamp = timestamp.replace(tzinfo=timezone.utc)
            fifteen_min = timedelta(minutes=15)
            return abs(self.now - timestamp) > fifteen_min
        return True

    def is_valid(self, secret_key):
        if self.is_expired():
            self.error_message = "signature is expired"
            return False
        signatures_equals = self._signature_equals(
            self._get_signature(secret_key), self.signature_in_header
        )
        if not signatures_equals:
            self.error_message = "signature invalid"
        return signatures_equals


class HmacSignature(BaseHmacSignature):
    def parse_request(self, request):
        self.timestamp = datetime.strftime(self.now, TIMESTAMP_FORMAT)
        self.headers_to_sign = {
            "host": request["host"],
            self.date_header: self.timestamp,
        }
        self.method = request["method"]
        self.path = request["path"]
        self.params = request["params"]
        if "json" in request and request["json"]:
            self.data = json.dumps(request["json"])
        else:
            self.data = request.get("data", "")
        self.host = request["host"]

    def generate_headers(self, credentials):
        self.access_key = credentials["access_key"]
        signature = self._get_signature(credentials["secret_key"])

        authorization_header = [
            f"{self.algorithm} Credential={self.access_key}/{self._get_credential_scope()}"
        ]
        authorization_header.append(
            f"SignedHeaders={self._get_signed_headers(self.headers_to_sign)}"
        )
        authorization_header.append("Signature=%s" % signature)
        return {
            "host": self.host,
            self.date_header: self.timestamp,
            "authorization": ", ".join(authorization_header),
        }


class HmacAuthBase(AuthBase):
    """Extend AuthBase from python requests with HMAC authentication"""

    def __init__(
        self,
        access_key,
        secret_key,
        service,
        region,
        service_key="dci2_request",
        algorithm="DCI2-HMAC-SHA256",
    ):
        self.access_key = access_key
        self.secret_key = secret_key
        self.service = service
        self.region = region
        self.service_key = service_key
        self.algorithm = algorithm

    def __call__(self, r):
        url = urlparse(r.url)
        signature = HmacSignature(
            {
                "service_name": self.service,
                "service_key": self.service_key,
                "region_name": self.region,
                "algorithm": self.algorithm,
            }
        ).add_request(
            {
                "method": r.method,
                "params": dict(parse_qsl(url.query)),
                "data": self.get_body(r.body),
                "host": url.netloc,
                "path": url.path,
            }
        )
        r.headers.update(
            signature.generate_headers(
                {
                    "access_key": self.access_key,
                    "secret_key": self.secret_key,
                },
            )
        )
        return r

    def get_body(self, body):
        if hasattr(body, "read"):
            c = body.read()
            body.seek(0)
            return c
        return body
